/*******************************************************************************
 * Western Digital (Paradise) PVGA1A chipset driver
 *******************************************************************************
 *
   Copyright (C) 1996, 1997 The Linux-GGI Project
   Copyright (C) 1997 by Brian S. Julin <bri@calyx.com>
 *
   Permission is hereby granted, free of charge, to any person obtaining a
   copy of this software and associated documentation files (the "Software"),
   to deal in the Software without restriction, including without limitation
   the rights to use, copy, modify, merge, publish, distribute, sublicense,
   and/or sell copies of the Software, and to permit persons to whom the
   Software is furnished to do so, subject to the following conditions:

   The above copyright notice and this permission notice shall be included in
   all copies or substantial portions of the Software.

   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
   IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
   FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
   THE AUTHOR(S) BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
   IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
   CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 *
 *******************************************************************************
 *
 * Version 0.3 Brian S. Julin
 * Minor: Set ramsize bits correctly
 *        Fixed PEL 9-dot font bug.
 *
 * Version 0.2 Brian S. Julin 
 * Minor: Bugfixes by Glen Re: 4BIT mode ATC setup
 *        Synced with changes in VGA Driver
 *
 * Version 0.1 Brian S. Julin
 * Major: Rewrote chipset_set_mode to work with BIOS default mode
 *        Cause ModeX to use special GRC[0E] Bit 1 -- fixes 640x400x8
 *        Added chipset detection
 *
 * Version 0.0 Brian S. Julin
 * Major: Copied IBM vga driver from GGI 0.0.9 12/13/97 Snapshot
 *
 *******************************************************************************
 * 
 * This code was derived from the following sources of information:
 * [1] Xfree86 VGA driver
 * [2] Zen of Graphics Programming (Michael Abrash)
 * [3] VGADOC Version 4b Finn Thoegersen
 */

#define _pvga1a_c

#include <kgi/maintainers.h>
#define	MAINTAINER	Brian_S_Julin
#define DRIVER_NAME	"PVGA1 driver"
#define DRIVER_REV	"$Revision: 1.6 $"

#define DEBUG_LEVEL	0

#include <kgi/module.h>
#include "WD/pvga1a_io.h"
#include "WD/wd_family.h"

#define CURSOR_HW	/* Hardware cursor */
#define TEXT_A000	/* Base of textmode framebuffer */
#define	POINTER_FONT	/* Font-hack mouse cursor in textmodes */
#undef SLOW_POINTER	/* #define this if your pointer flickers */

kgiu16 CRTI, CRTD, INS1r;
static char *wd_chipset_names[WD_HOWMANY_CHIPSETS] = { WD_VERBOSE_NAMES };

/* For insmodline options */
char *chipset;

struct kgi_wd_chipset chipinfo =
{
  { "Western Digital",	          /* Manufacturer		    */
    "PVGA1A",		          /* Model	        	    */
    GM_TEXT16 | GM_1BIT |	  /* Available modes	            */
    GM_4BIT | GM_8BIT,
    0,			          /* Flags	         	    */
    { 1600, 1200  },	          /* Maximum resolution	            */
    0x00000,		          /* RAM gets detected below        */
    { 25175000, 28322000 },	  /* DCLK limits		    */
  },
  /* The rest is additional information in the WD series of drivers */
  0,                              /* Default features: Nonbanking RAM   */
  CHIPSET_NOT_WD                  /* Chipset type                   */
};

/* This table holds the default values for the VGA registers.  A copy	*/
/* of it will be filled in as appropriate during mode programming in	*/
/* kgim_chipset_set_mode().  These values are based on BIOS mode 0002h  */
/* as read from a wd90c90 ISA card. */
struct statetable default_state =
{
  0x67,                                                    /* MISC */
  
  {0x03, 0x00, 0x03, 0x00, 0x02},                          /*  SEQ */
  
  {0x5f, 0x4f, 0x50, 0x82, 0x55, 0x81, 0xbf, 0x1f, 
   0x00, 0x4f, 0x0d, 0x0e, 0x00, 0x00, 0x00, 0x00, 
   0x9c, 0x0e, 0x8f, 0x28, 0x1f, 0x96, 0xb9, 0xa3, 
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 
   0xff, 0xff, 0xff, 0xff, 0xff},                          /*  CRT */
  
  {0x00, 0x00, 0x00, 0x00, 0x00, 0x10, 0x0e, 0x00, 
   0xff, 0x00, 0x00, 0x46, 0x00, 0x00, 0x00, 0x95 },       /*  GRC */
  
  {0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x14, 0x07, 
   0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f, 
   0x0c, 0x00, 0x0f, 0x00, 0x00 }                          /*  ATC */
};

/* Hack until clockchip stuff is done */
int long pvga1a_clockset[0x10] =
{  
   25175000, 28322000, 0, 0,
   25175000, 28322000, 0, 0,
   25175000, 28322000, 0, 0,
   25175000, 28322000, 0, 0
};

struct statetable reset_state;
kgi_modereq default_kgi_modereq;
struct kgi_mode default_kgi_mode;

static unsigned long fb_text_len;
static unsigned long fb_text;
static unsigned long fb_graph_len;
static unsigned long fb_graph;

static int detect_PVGA1A(void)
{
  int err;

  /* Get essential PIO registers */
  if ((err = io_check_region(&wd_wakeup))) {
    printk("Wakeup Register Not available -- won't sleep card\n");
    printk("Checking for already alert PVGA1A...\n");
  }
  else {
    io_claim_region(&wd_wakeup);
    printk("Got sleep register\n");
  }
  
  /* Cause IO base to be determined */
  /* IMPORTANT: Must be done BEFORE accesing CRT */
  outbMISC(inbMISC());
  kgi_chipset_wd_detect(&chipinfo);
  
  if (chipinfo.chiptype == 0 ) 
    {
      printk("Cannot find WD chipset, aborting\n");
      return -E(CHIPSET,NOSUP);
    };
  if ((chipinfo.chiptype == CHIPSET_UNKNOWN_WD) &&
      ((chipset == NULL) || (strstr(chipset,"pvga1a") == NULL)))
    {
      printk("Unknown WD chipset, aborting\n");
      return -E(CHIPSET,NOSUP);
    };

  if (! chipinfo.revstring[0]) {
    strcpy(chipinfo.revstring, "N/A\0");
  } 
  printk("Detected Chipset %s (Revision info: %s)\n", 
	 wd_chipset_names[chipinfo.chiptype], chipinfo.revstring);
  
  if (! (chipinfo.chiptype == CHIPSET_PVGA1A))
		if ((chipset != NULL) && (strstr(chipset,"pvga1a") != NULL)) {
			printk("User Override: treating chipset as if it were PVGA1A\n");
		}
		else {
			printk("This driver for PVGA1 chipsets only, aborting\n");
			return -E(CHIPSET,NOSUP);
		};
  
  kgi_chipset_wd_findram(&chipinfo);
  printk("Detected %dKb of RAM\n", chipinfo.chipset.ramsize / 1024);

#ifdef TEXT_A000
	fb_text		= 0xA0000;
	fb_text_len	= 0x10000;	/* 64k framebuffer */
#else
	fb_text		= 0xB8000;
	fb_text_len	= 0x8000;
#endif /* TEXT_A000 */

	fb_graph_len = 0x20000;	/* We have only 64K linear graphics */
	fb_graph = fb_text;

	return EOK;
}

/* GGI interface */

/* Mode specific frame buffers */

#ifdef POINTER_FONT
# include "../IBM/vga_text16_enh.inc"
#else
# include "../IBM/vga_text16.inc"
#endif

#include "../IBM/vga_fb.inc"

#ifdef DEBUG_FB
# include "../IBM/vga_debug_fb.inc"
#endif

static void chipset_setup_dpy(struct kgi_display *dpy, struct kgi_mode *mode)
{
	switch (mode->request.graphtype) {
		case KGIGT_TEXT16:
		dpy->set_font		= text16_set_font;
		dpy->set_origin		= text16_set_origin;
		dpy->set_split		= text16_set_split;
		dpy->set_frame		= text16_set_frame;
		dpy->show_cursor	= text16_show_cursor;
		dpy->hide_cursor	= text16_hide_cursor;
		dpy->undo_cursor	= text16_undo_cursor;
		dpy->show_pointer	= text16_show_pointer;
		dpy->hide_pointer	= text16_hide_pointer;
		dpy->undo_pointer	= text16_undo_pointer;
		dpy->fb.dpy		= dpy;
		dpy->fb.size.x 		= mode->request.virt.x;
		dpy->fb.size.y		= mode->request.virt.y;
		dpy->fb.bpp		= MAP_16_BPP;
		dpy->fb.pbuf		= phys_to_virt(fb_text);
		dpy->fb.zbuf		= NULL;
		dpy->fb.ops		= &text16_color_io;
 		KGIM_PRIVATE(dpy, driver) = 
		  text16_data_realloc(KGIM_PRIVATE(dpy,driver));
 		if (KGIM_PRIVATE(dpy, driver) == NULL)
			INTERNAL_ERROR;
		break;
		default:
		dpy->set_font		= NULL;
		dpy->set_origin		= vgafb_set_origin;
		dpy->set_split		= vgafb_set_split;
		dpy->set_frame		= vgafb_set_frame;
		dpy->show_cursor	= NULL;
		dpy->hide_cursor	= NULL;
		dpy->undo_cursor	= NULL;
		dpy->show_pointer	= NULL;
		dpy->hide_pointer	= NULL;
		dpy->undo_pointer	= NULL;
		dpy->fb.dpy		= dpy;
		dpy->fb.size.x		= 0;
		dpy->fb.size.y		= 0;
		dpy->fb.bpp		= 0;
		dpy->fb.pbuf		= NULL;
		dpy->fb.zbuf		= NULL;
		dpy->fb.ops		= NULL;
		kgim_graphics_set_mode(dpy,mode);
	}
	memcpy(&(dpy->mode), mode, sizeof(struct kgi_mode));
}

/* Takes a mode and a bits-per-pixel value */
static void chipset_setup_mode(struct kgi_mode *mode)
{
	int bpp=0;

	/* By default, setup for graphics modes */

	switch (mode->request.graphtype) {
		case KGIGT_TEXT16:
		mode->fb	= (mem_vaddr)fb_text;
 		mode->fb_size	= fb_text_len;
		mode->fsize	= 
			mode->request.virt.x * mode->request.virt.y * 2;
		mode->mem	= &text16_color_mem;
		mode->io	= &text16_color_io;
		return; 	/* Done setting up */
		case KGIGT_1BIT:
		bpp = 1;
		break;
		case KGIGT_4BIT:
		bpp = 4;
		break;
		case KGIGT_8BIT:
		bpp = 8;
		break;
		default:
		INTERNAL_ERROR;
		return;
	}
	
	mode->fb	= (mem_vaddr)fb_graph;
	mode->fb_size	= fb_graph_len;
	mode->fsize	= mode->request.virt.x * mode->request.virt.y * 8 / bpp;
	mode->mem	= NULL;
	mode->io	= NULL;
}

static int chipset_get_reset_mode(struct kgi_display *dpy)
{
	kgi_graphtype gmode = KGIGT_TEXT16;
	int magx = 1, magy = 1;
	int DCLKdiv = 8;
	int dotclock;
	struct kgi_timing tx,ty;

	save_state(&reset_state);

	if (reset_state.ATC[0x10] & AR10_GRAPHICS) {
		switch (reset_state.GRC[0x05] & GR5_WRITEMODE3) {
			case GR5_WRITEMODE0:
			gmode = KGIGT_8BIT; /* 8 bit mode */
			break;
			case GR5_WRITEMODE1:
			gmode = KGIGT_1BIT; /* 1 bit mode */
			break;
			case GR5_WRITEMODE3:
			gmode = KGIGT_4BIT; /* 4 bit mode */
			break;
			case GR5_WRITEMODE2:
			break;
		}
		magy = (reset_state.CRT[0x09] & CR09_CHARHEIGHT_MASK) + 1;
	} else { /* Text mode */
		gmode = KGIGT_TEXT16;
		switch (reset_state.GRC[GRC_VID_CONF] & GR0C_CHARDOT_MASK)
		  {
		  case GR0C_8OR9DOT_CHARS :
		    DCLKdiv = (reset_state.SEQ[SEQ_CLOCK] & SR01_8DOT_CHCLK) 
		      ? 8 : 9;
		    break;
		  case GR0C_7DOT_CHARS :
		    DCLKdiv = 7;
		    break;
		  case GR0C_9DOT_CHARS :
		    DCLKdiv = 9;
		    break;
		  case GR0C_6OR10DOT_CHARS :
		    DCLKdiv = 10;
		    break;
		  default :
		    INTERNAL_ERROR;
		  };		
	}
    
	/* Get horizontal timing */
	tx.width	= (reset_state.CRT[0x01]+1) * DCLKdiv;
	tx.blankstart	= reset_state.CRT[0x02] * DCLKdiv;
	tx.syncstart	= reset_state.CRT[0x04] * DCLKdiv;
	tx.syncend	= ((reset_state.CRT[0x05]) & CR05_HSE_MASK) * DCLKdiv;
	while (tx.syncend < tx.syncstart)
		tx.syncend += 0x20 * DCLKdiv;
	tx.blankend	= (
		(reset_state.CRT[0x03] & CR03_HBE_MASK) |
		((reset_state.CRT[0x05] & CR05_HBLANKEND_B5) ? 0x20 : 0) |
		(reset_state.CRT[0x01] & 0xC0)) * DCLKdiv;
	tx.total	= (reset_state.CRT[0x00]+5) * DCLKdiv;
	tx.polarity	= (reset_state.MISC & MISC_NEG_HSYNC) != 0;

	/* Get vertical timing */
	ty.width	= 1 + (
		reset_state.CRT[0x12] | 
		((reset_state.CRT[0x07] & CR07_VDISPLAYEND_B8) ? 0x100 : 0) |
		((reset_state.CRT[0x07] & CR07_VDISPLAYEND_B9) ? 0x200 : 0));
	ty.blankstart	= 1 + (
		reset_state.CRT[0x15] |
		((reset_state.CRT[0x07] & CR07_VBLANKSTART_B8) ? 0x100 : 0) |
		((reset_state.CRT[0x09] & CR09_VBLANKSTART_B9) ? 0x200 : 0));
	ty.syncstart	= (
		reset_state.CRT[0x10] |                
		((reset_state.CRT[0x07] & CR07_VSYNCSTART_B8) ? 0x100 : 0) |
		((reset_state.CRT[0x07] & CR07_VSYNCSTART_B9) ? 0x200 : 0));
	ty.syncend	= (reset_state.CRT[0x11] & CR11_VSYNCEND_MASK) 
		| (reset_state.CRT[0x12] & ~CR11_VSYNCEND_MASK);
	while (ty.syncend < ty.syncstart)
		ty.syncend += 0x10;
	ty.blankend	= (reset_state.CRT[0x16] | ((ty.width - 1) & ~0xFF)) + 1;
	ty.total = 2 + (
		reset_state.CRT[0x06] |
		((reset_state.CRT[0x07] & CR07_VTOTAL_B8) ? 0x100 : 0) |
		((reset_state.CRT[0x07] & CR07_VTOTAL_B9) ? 0x200 : 0));
	ty.polarity = (reset_state.MISC & MISC_NEG_VSYNC) != 0;

	dotclock = pvga1a_clockset[
		((reset_state.MISC & MISC_CLOCK_SEL_B0) ? 1 : 0) +
	  ((reset_state.MISC & MISC_CLOCK_SEL_B1) ? 2 : 0) +
	  ((reset_state.GRC[GRC_VID_CONF] & GR0C_CLOCK_SEL_B2) ? 4 : 0)];
	
	/* Set all the mode info that we've just gathered */
	default_kgi_modereq.frames		= 1;
	default_kgi_modereq.dpp.x		= DCLKdiv;
	default_kgi_modereq.dpp.y		= (
		reset_state.CRT[0x09] & CR09_CHARHEIGHT_MASK) + 1;
	default_kgi_modereq.visible.x	= 
		default_kgi_modereq.virt.x	= tx.width / default_kgi_modereq.dpp.x;
	default_kgi_modereq.visible.y	= 
		default_kgi_modereq.virt.y	= ty.width / default_kgi_modereq.dpp.y;
	default_kgi_modereq.size.x		= 0;
	default_kgi_modereq.size.y		= 0;
	default_kgi_modereq.graphtype	= gmode;
	if (dpy->defmode == NULL)
		dpy->defmode = &default_kgi_modereq;
	else
		(*dpy->defmode) = default_kgi_modereq;

	dpy->mode.request	= default_kgi_modereq;
	dpy->mode.x		= tx;
	dpy->mode.y		= ty;
	dpy->mode.magnify.x	= magx;
	dpy->mode.magnify.y	= magy;
	dpy->mode.dclk		= dotclock;
	dpy->mode.clk.mul	= 1;
	dpy->mode.clk.div	= 1;
	dpy->mode.lclk.mul	= 1;
	dpy->mode.lclk.div	= 1;
	dpy->mode.chipset	= 0;
	
	default_kgi_mode = dpy->mode;

	chipset_setup_mode(&default_kgi_mode);
	chipset_setup_dpy(dpy, &(default_kgi_mode));

	return EOK;
}
   
static int chipset_init(struct kgi_display *dpy)
{
	int err;

	NOTICE(DRIVER_NAME " rev " DRIVER_REV);

	if (detect_PVGA1A() == EOK) {
		if ((err = mem_check_region(&wd_mem_graph)))
			return err;
#ifndef TEXT_A000
		if ((err = mem_check_region(&wd_mem_text)))
			return err;
#endif

		mem_claim_region(&wd_mem_graph);
#ifndef TEXT_A000
		mem_claim_region(&wd_mem_text);
#endif

		text16_data_init();
		return chipset_get_reset_mode(dpy);
	}

	return -E(CHIPSET,FAILED);
}

static void chipset_reset(struct kgi_display *dpy)
{
	restore_state(&reset_state);
	if (dpy) {
		dpy->mode = default_kgi_mode;
		(*dpy->defmode) = default_kgi_modereq;
	}
}

static void chipset_done(void)
{
	chipset_reset(NULL);

	io_free_region(&wd_wakeup);
	mem_free_region(&wd_mem_graph);
#ifndef TEXT_A000
	mem_free_region(&wd_mem_text);
#endif
	
}

static void adjust_timing(struct kgi_mode *mode)
{
	register int foo = (mode->request.graphtype == KGIGT_TEXT16)
		? mode->request.dpp.x : 8;

#define ADJUST(X) mode->x.X -= (mode->x.X % foo)

	ADJUST(width);
	ADJUST(blankstart);
	ADJUST(syncstart);
	ADJUST(syncend);
	ADJUST(blankend);
	if (mode->x.total % foo)
		mode->x.total += foo - (mode->x.total % foo);

	if (((mode->x.syncend / foo) & 0x3F) == 
			((mode->x.syncstart / foo) & 0x3F))
		mode->x.syncend -= foo;
	if (((mode->x.blankend / foo) & 0x7F) == 
			((mode->x.blankstart / foo) & 0x7F))
		mode->x.blankend -= foo;

	if ((mode->y.syncend & 0x0F) == (mode->y.syncstart & 0x0F))
		mode->y.syncend--;
	if ((mode->y.blankend & 0xFF) == (mode->y.blankstart & 0xFF))
		mode->y.blankend--;

#undef ADJUST
}

static int chipset_check_mode(
	struct kgi_display *dpy, 
	struct kgi_mode *mode,
	enum kgi_tm_cmd cmd)
{
	int bpp = 0;
	int dppx, dppy;
	dppx = mode->request.dpp.x;
	dppy = mode->request.dpp.y;

	switch (mode->request.graphtype) {
		case KGIGT_8BIT:
		bpp = 8;	/* 1 Bpp */
		break; 	       
		case KGIGT_1BIT:
		case KGIGT_4BIT:
		bpp = 4;	/* 0.5 Bpp */
		break;	
		case KGIGT_TEXT16:
		break;		/* Not used in textmode */
		default:
		DEBUG0("graphtype %i not supported", mode->request.graphtype);
		return -E(CHIPSET,FAILED);
	}

#define XVIRTUAL (mode->request.virt.x)
#define YVIRTUAL (mode->request.virt.y)
#define XVISIBLE (mode->request.visible.x)
#define YVISIBLE (mode->request.visible.y)

	switch (cmd) {
		case CMD_PROPOSE:
		if (mode->request.graphtype == KGIGT_TEXT16) {

			mode->magnify.x = mode->magnify.y = 1;
			if ((dppx < 7) || (10 < dppx)) {
				DEBUG0("Invalid char width (%i)", dppx);
				return -EINVAL;
			}

			if (XVIRTUAL == 0 && YVIRTUAL == 0)
				return -EINVAL;

			if (XVIRTUAL == 0) 
				XVIRTUAL = fb_text_len / YVIRTUAL / 2;
			
			if (YVIRTUAL == 0)
				YVIRTUAL = fb_text_len / XVIRTUAL / 2;

			if (2 * XVIRTUAL * YVIRTUAL > fb_text_len) {
				NOTICE("Not enough text memory");
				return -ENOMEM;
			}

			mode->x.width = mode->request.visible.x * 
				mode->request.dpp.x * mode->magnify.x;
			mode->y.width = mode->request.visible.y * 
				mode->request.dpp.y * mode->magnify.y;

		} else {	/* mode->request.graphtype == KGIGT_TEXT16 */

			if (XVIRTUAL == 0 && YVIRTUAL == 0)
				return -EINVAL;

			if (XVIRTUAL == 0)
				XVIRTUAL = fb_graph_len / YVIRTUAL;

			if (YVIRTUAL == 0)
				YVIRTUAL = fb_graph_len / XVIRTUAL;

			mode->magnify.x = 1;
			mode->magnify.y = 1;

			while ((XVISIBLE * mode->magnify.x) < 640) 
				mode->magnify.x++;

			while ((YVISIBLE * mode->magnify.y) < 350) 
				mode->magnify.y++;
			
			/* Can't zoom more than this */
			if (mode->magnify.y > 2)
				return -EINVAL;
			
				/* Can't do more than doublescan */
			if (mode->magnify.x > 2)
				return -EINVAL;
			
			if ((XVISIBLE * YVISIBLE * bpp / 8) > 
					chipinfo.chipset.ramsize) {
				DEBUG0(
					"Not enough memory to do mode %ix%i at %i bpp",
					XVISIBLE, YVISIBLE, bpp);
				return -ENOMEM;
			} 
			
			if ((XVIRTUAL * YVIRTUAL * bpp / 8) > 
					chipinfo.chipset.ramsize) {
				DEBUG0("Lack of memory forced virtual = visible\n");
				XVIRTUAL = XVISIBLE;
				YVIRTUAL = YVISIBLE;
			}
			mode->request.visible.x &= 0x7FFFFFF8;
			mode->request.visible.y &= 0x7FFFFFF8;
			mode->x.width = 
				mode->request.visible.x * mode->magnify.x;
			mode->y.width = 
				mode->request.visible.y * mode->magnify.y;
		}

		
		return EOK;
		
		case CMD_LOWER:

		if (mode->dclk < chipinfo.chipset.dclk.min) {
			DEBUG1("DCLK too low (%i)", mode->dclk);
			return -E(CHIPSET,FAILED);
		}
		mode->dclk = (mode->dclk > chipinfo.chipset.dclk.max)
			? chipinfo.chipset.dclk.max : mode->dclk;

		adjust_timing(mode);
		return EOK;

		case CMD_RAISE:

		if (mode->dclk > chipinfo.chipset.dclk.max) {
			DEBUG1("DCLK too high (%i)", mode->dclk);
			return -E(CHIPSET,FAILED);
		}
		mode->dclk = (mode->dclk < chipinfo.chipset.dclk.min)
			? chipinfo.chipset.dclk.min : mode->dclk;

		adjust_timing(mode);
		return EOK;

		case CMD_CHECK:
		if (mode->lclk.mul != 1) {
			DEBUG1("Invalid lclk multiplier %d", mode->lclk.mul);
			return -E(CHIPSET,FAILED);
		}
		switch (mode->lclk.div) {
			case 1:
			break;
			default:
			DEBUG1("Invalid CLK to LCLK divider %d", mode->lclk.div);
			return -E(CHIPSET,FAILED);
	    }
		if (mode->request.graphtype == KGIGT_TEXT16) {
			mode->flags |= FL_CHIP_COLORTEXT;
			if ((mode->magnify.x != 1) || (mode->magnify.y != 1)
			    || (dppx < 7) || (10 < dppx) || (dppy < 1)
			    || (dppy > 32) ||
			    (2 * XVIRTUAL * YVIRTUAL > fb_text_len)
			    || (mode->dclk > chipinfo.chipset.dclk.max)
			    || (mode->dclk < chipinfo.chipset.dclk.min))
			{
				NOTICE1("HW check failed");
				return -E(CHIPSET,FAILED);
			}
		} else {
			switch (XVIRTUAL) {
				case 320:
				case 360:
				case 640:
				case 720:
				break;
				default:
				return -E(CHIPSET,FAILED);
			}

			if (((XVIRTUAL * YVIRTUAL * bpp / 8) 
					 > chipinfo.chipset.ramsize)
			    || (mode->magnify.x > 2) || (mode->magnify.y > 2)
			    || (mode->clk.mul != 1) || (mode->clk.div != 1)
			    || (mode->dclk < chipinfo.chipset.dclk.min)
			    || (chipinfo.chipset.dclk.max < mode->dclk)
			    || (mode->magnify.x == 2 && mode->request.graphtype != KGIGT_8BIT))
				return -E(CHIPSET,FAILED);
		}

#define TM(X) ((chipinfo.chipset.maxdots.x < mode->x.X) || (chipinfo.chipset.maxdots.y < mode->y.X))
		if (TM(width) || TM(blankstart) || TM(syncstart)
		    || TM(syncend) || TM(blankend) || TM(total)) {
			NOTICE1("Timing check failed");
			DEBUG1("Maxdots are (%d,%d)", chipinfo.chipset.maxdots.x,
				chipinfo.chipset.maxdots.y);
			return -E(CHIPSET,FAILED);
		}
#undef TM

		chipset_setup_mode(mode);
		return EOK;

		default:
		INTERNAL_ERROR;
		return -E(CHIPSET,FAILED);
	}
	
#undef XVIRTUAL
#undef YVIRTUAL
#undef XVISIBLE
#undef YVISIBLE
}

/* This should probably be changed to an inline */
#define KGI_POKEBITS(VAL,MASK,REG) REG = ((REG & ~MASK) | (MASK & VAL))

static void chipset_set_mode(struct kgi_display *dpy,struct kgi_mode *mode)
{
	struct statetable setup = default_state;
	register int foo;
	int DCLKdiv = 8;
	int dppx,dppy;
	dppx = mode->request.dpp.x;
	dppy = mode->request.dpp.y;

	/* Set up bits per pixel and logical screen width */

	switch (mode->request.graphtype) {
		case KGIGT_TEXT16:
		foo = 2 * mode->request.virt.x;
		foo /= 4;	/* Due to the addressing mode used */
		break;
		case KGIGT_1BIT:
		case KGIGT_4BIT:
		foo = mode->request.virt.x >> 4;
		break;
		case KGIGT_8BIT:
		foo = mode->request.virt.x >> 3;
		break;
		default:
		INTERNAL_ERROR;
		return;
	}

	setup.CRT[CRT_LOGICALWIDTH] = foo & 0xFF;

	if (mode->request.graphtype == KGIGT_TEXT16) {
		setup.CRT[CRT_MODE] = 
		  CR17_CGA_ADDR_WRAP | CR17_ENABLE_SYNC | 
		  CR17_CGA_BANKING | CR17_HGC_BANKING;
		
		setup.CRT[CRT_UNDERLINE] |= CR14_UNDERLINE_MASK;
		
		KGI_POKEBITS(
			(SR02_PLANE0 | SR02_PLANE1),
			SR02_PLANE_MASK, setup.SEQ[SEQ_WPLANE]);

		setup.GRC[GRC_MODECONTROL] |= GR5_ODD_EVEN;
		setup.GRC[GRC_MMAP_MODE] |= GR6_CHAIN_OE;
#if defined(TEXT_A000)
		KGI_POKEBITS(
			GR6_MAP_A0_64, GR6_MAP_MASK, setup.GRC[GRC_MMAP_MODE]);
#else
		KGI_POKEBITS(
			GR6_MAP_B8_32, GR6_MAP_MASK, setup.GRC[GRC_MMAP_MODE]);
#endif /* TEXT_A000 */

			/* Set bank offset registers to 0 -- shouldn't
				 graphics driver have done this? */
		setup.GRC[GRC_PR0A] = 0;
		setup.GRC[GRC_PR0B] = 0; 

		setup.SEQ[SEQ_MEMMODE] |= SR04_256K_ACCESS;

		setup.ATC[ATC_MODECONTROL] &= ~ 
		  (AR10_GRAPHICS | AR10_MONOMODE | 
		   0x10 /* reserved (0) */ | AR10_8BIT_MODE | AR10_ENB_BLINK);
		setup.ATC[ATC_MODECONTROL] |= (AR10_TOP_PAN | AR10_SELECT_B54);
		setup.ATC[ATC_BORDER] = 0;
		setup.ATC[ATC_HPAN] = 0;
		setup.ATC[ATC_PIXELPAD] &= ~ AR14_DATA_MASK;

		switch ((DCLKdiv = dppx)) {
		case 7:
		  KGI_POKEBITS(
				GR0C_7DOT_CHARS, GR0C_CHARDOT_MASK,
				setup.GRC[GRC_VID_CONF]);
		  break;
		case 8:
		  setup.SEQ[SEQ_CLOCK] |= SR01_8DOT_CHCLK;
		  KGI_POKEBITS(
				GR0C_8OR9DOT_CHARS, 
				GR0C_CHARDOT_MASK,
				setup.GRC[GRC_VID_CONF]);
		  setup.ATC[ATC_MODECONTROL] &= ~AR10_9DOT_CHARS;
		  break;
		case 9:
		  setup.SEQ[SEQ_CLOCK] &= ~SR01_8DOT_CHCLK;
		  KGI_POKEBITS(
				AR10_9DOT_CHARS,
				(AR10_9DOT_CHARS | AR10_TOP_PAN),
				setup.ATC[ATC_MODECONTROL]);
		  KGI_POKEBITS(
				GR0C_8OR9DOT_CHARS, 
				GR0C_CHARDOT_MASK,
				setup.GRC[GRC_VID_CONF]);
		  break;
		case 10:
		  KGI_POKEBITS(
				GR0C_6OR10DOT_CHARS, GR0C_CHARDOT_MASK,
				setup.GRC[GRC_VID_CONF]);
		  break;
		default:
		  INTERNAL_ERROR;
		  return;
		}
		KGI_POKEBITS(
			(dppy - 1), CR09_CHARHEIGHT_MASK,
			setup.CRT[CRT_CHARHEIGHT]);

		KGI_POKEBITS(
			((dppx == 9) ? 0x08 : 0x00), 
			AR13_DATA_MASK, setup.ATC[ATC_HPAN]); 

	} else {	/* mode->request.graphtype == KGIGT_TEXT16 */
		DCLKdiv = 8;

		setup.CRT[CRT_MODE] = 
		  CR17_CGA_BANKING | CR17_HGC_BANKING | 
			CR17_ENABLE_SYNC | CR17_CGA_ADDR_WRAP;

		for (foo = 0; foo < 16; foo++)
			setup.ATC[foo] = default_state.ATC[foo];

		setup.SEQ[SEQ_WPLANE] |= SR02_PLANE_MASK;
		setup.ATC[ATC_PLANEENABLE] |= AR12_PLANEMASK;

		KGI_POKEBITS(
			(SR04_256K_ACCESS | SR04_NO_ODDEVEN),
			0x0f, setup.SEQ[SEQ_MEMMODE]);

		setup.GRC[GRC_MODECONTROL] &= ~GR5_ODD_EVEN;
		KGI_POKEBITS(
			(GR6_GRAPHMODE | GR6_MAP_A0_64),
			(GR6_GRAPHMODE | GR6_MAP_MASK | GR6_CHAIN_OE),
			setup.GRC[GRC_MMAP_MODE]);

		setup.CRT[CRT_HBLANKEND] |= CR03_IS_VGA;
		setup.CRT[CRT_MODE] |= CR17_BYTE_MODE;

		KGI_POKEBITS(
			(mode->magnify.y - 1), CR09_CHARHEIGHT_MASK,
			setup.CRT[CRT_CHARHEIGHT]);

		KGI_POKEBITS(
			GR0C_8OR9DOT_CHARS,
			GR0C_CHARDOT_MASK,
			setup.GRC[GRC_VID_CONF]);
		setup.SEQ[SEQ_CLOCK] |= SR01_8DOT_CHCLK;

		setup.ATC[ATC_MODECONTROL] &= ~ 
		  (AR10_MONOMODE | AR10_ENB_BLINK | AR10_8BIT_MODE |
		   0x10 /* reserved (0) */ | AR10_SELECT_B54);
		setup.ATC[ATC_MODECONTROL] |=
		  (AR10_GRAPHICS | AR10_TOP_PAN);

		switch (mode->request.graphtype) {
		case KGIGT_1BIT:
		  /* Enable plane 0 for CPU */
		  KGI_POKEBITS(
				0x0e, GR1_DATA_MASK,
				setup.GRC[GRC_ENABLE]);
		  KGI_POKEBITS(
				0x00, GR4_PLANEMASK,
				setup.GRC[GRC_READPLANE]);
		  KGI_POKEBITS(
				GR5_WRITEMODE0, 0x03,
				setup.GRC[GRC_MODECONTROL]); 
		  break;
		case KGIGT_4BIT:
		  /* Enable all planes for set/reset (color banking) */
		  KGI_POKEBITS(
				0x0f, GR1_DATA_MASK,
				setup.GRC[GRC_ENABLE]);
		  KGI_POKEBITS(
				GR5_WRITEMODE3, 0x03,
				setup.GRC[GRC_MODECONTROL]); 
		  KGI_POKEBITS(
				0x0F, AR12_PLANEMASK, 
				setup.ATC[ATC_PLANEENABLE]);
		  break;

		case KGIGT_8BIT:

		  /* Planar (Banked) */
		  setup.SEQ[SEQ_MEMMODE] |= SR04_CHAINED;

		  setup.CRT[CRT_UNDERLINE] |= CR14_DOUBLEWORDMODE;
		  KGI_POKEBITS(
				((mode->magnify.x == 1) 
				 ? GR0E_SHIFT_256 : 0),
				GR0E_SHIFT_256, setup.GRC[GRC_VID_CTL]);
		  
		  setup.ATC[ATC_MODECONTROL] |= AR10_8BIT_MODE;
		  setup.ATC[ATC_BORDER] = 0;
		  setup.ATC[ATC_HPAN] = 0;
		  setup.ATC[ATC_PIXELPAD] &= ~ AR14_DATA_MASK;
		  
		  setup.GRC[GRC_MODECONTROL] |= GR5_8BIT_COLOR;
		  setup.GRC[GRC_BITMASK] = 0xff;	/* Enable all bits */
		  
		  KGI_POKEBITS(
				0x00, GR1_DATA_MASK,
				setup.GRC[GRC_ENABLE]);
		  KGI_POKEBITS(
				GR5_WRITEMODE0, 0x03,
				setup.GRC[GRC_MODECONTROL]);
		  
		  break;
		default:
		  INTERNAL_ERROR;
		  return;
		}
	}

		/* "Available Memory" registers are writable, so set to correct 
			 values, else chipset will behave as if less RAM is present */
	switch (chipinfo.chipset.ramsize) {
	case (256 * 1024):
	  KGI_POKEBITS(
			GR0B_256K_BANKED, GR0B_RAMSIZE, 
			setup.GRC[GRC_MEMORY]);
	  break;
	case (512 * 1024):
	  KGI_POKEBITS(
			GR0B_512K_BANKED, GR0B_RAMSIZE, 
			setup.GRC[GRC_MEMORY]);
	  break;
	case (1024 * 1024):
	default:  /* Accomodate wd90c33 chips using this driver */
	  KGI_POKEBITS(
			GR0B_1M_BANKED, GR0B_RAMSIZE, 
			setup.GRC[GRC_MEMORY]);
	  break;
	};

		/* Set up horizontal timing */
	
		/* Horizontal display end */
	foo = (mode->x.width / DCLKdiv) - 1;
	setup.CRT[CRT_HDISPLAYEND] = foo & 0xFF;

		/* Horizontal blank start */
	foo = (mode->x.blankstart / DCLKdiv);	
	setup.CRT[CRT_HBLANKSTART] = foo & 0xFF;

		/* Horizontal sync start  */
	foo = (mode->x.syncstart / DCLKdiv);
	setup.CRT[CRT_HSYNCSTART] = foo & 0xFF;
	
		/* Horizontal sync end    */
	foo = (mode->x.syncend / DCLKdiv);	
	KGI_POKEBITS(
		foo, CR05_HSE_MASK, 
		setup.CRT[CRT_HSYNCEND]);

		/* Horizontal blank end   */
	foo = (mode->x.blankend / DCLKdiv);
	KGI_POKEBITS(
		foo, CR03_HBE_MASK, 
		setup.CRT[CRT_HBLANKEND]); 
	KGI_POKEBITS(
		((foo & 0x20) ? CR05_HBLANKEND_B5 : 0),
		CR05_HBLANKEND_B5, setup.CRT[CRT_HSYNCEND]); 

		/* Horizontal total       */
	foo = (mode->x.total / DCLKdiv) - 5;
	setup.CRT[CRT_HTOTAL] = foo & 0xFF;
	
		/* Set up vertical timing */
	
		/* Vertical display end   */
	foo = mode->y.width - 1;		
	setup.CRT[CRT_VDISPLAYEND] = foo & 0xFF;
	KGI_POKEBITS(
		(((foo & 0x100) ? CR07_VDISPLAYEND_B8 : 0) |
		 ((foo & 0x200) ? CR07_VDISPLAYEND_B9 : 0)),
		(CR07_VDISPLAYEND_B8 | CR07_VDISPLAYEND_B9), 
		setup.CRT[CRT_OVERFLOW]); 
	
	foo = mode->y.blankstart - 1;		/* Vertical blank start   */
	setup.CRT[CRT_VBLANKSTART] = foo & 0xFF;
	KGI_POKEBITS(
		((foo & 0x100) ? CR07_VBLANKSTART_B8 : 0),		    
		CR07_VBLANKSTART_B8, setup.CRT[CRT_OVERFLOW]); 
	KGI_POKEBITS(
		((foo & 0x200) ? CR09_VBLANKSTART_B9 : 0),
		CR09_VBLANKSTART_B9, setup.CRT[CRT_CHARHEIGHT]); 
	
	foo = mode->y.syncstart;		/* Vertical sync start    */
	setup.CRT[CRT_VSYNCSTART] = foo & 0xFF;
	KGI_POKEBITS(
		(((foo & 0x100) ? CR07_VSYNCSTART_B8 : 0) |
		 ((foo & 0x200) ? CR07_VSYNCSTART_B9 : 0)),
		(CR07_VSYNCSTART_B8 | CR07_VSYNCSTART_B9), 
		setup.CRT[CRT_OVERFLOW]); 

		/* Vertical sync end      */
	foo = mode->y.syncend;
	KGI_POKEBITS(foo, CR11_VSYNCEND_MASK, setup.CRT[0x11]);

		/* Vertical blank end     */
	foo = mode->y.blankend + 1;		
	setup.CRT[CRT_VBLANKEND] = foo & 0xFF;

		/* Vertical total         */
	foo = mode->y.total - 2;
	setup.CRT[CRT_VTOTAL] = foo & 0xFF;
	KGI_POKEBITS(
		(((foo & 0x100) ? CR07_VTOTAL_B8 : 0) |
		 ((foo & 0x200) ? CR07_VTOTAL_B9 : 0)),
		(CR07_VTOTAL_B8 | CR07_VTOTAL_B9), 
		setup.CRT[CRT_OVERFLOW]); 
	
	KGI_POKEBITS(
		(((mode->x.polarity > 0) ? 0 : MISC_NEG_HSYNC) |
		 ((mode->y.polarity > 0) ? 0 : MISC_NEG_VSYNC)),
		(MISC_NEG_HSYNC | MISC_NEG_VSYNC),
		setup.MISC);
	
		/* Set splitline to 0 */
	
	setup.CRT[CRT_LINECOMPARE] = 0;
	KGI_POKEBITS(
		0, CR07_LINECOMPARE_B8, 
		setup.CRT[CRT_OVERFLOW]);
	KGI_POKEBITS(
		0, CR09_LINECOMPARE_B9, 
		setup.CRT[CRT_CHARHEIGHT]);

	/* Ok.  We'll do the actual hardware programming now. */

	kgim_monitor_set_mode(dpy, mode);
	kgim_ramdac_set_mode(dpy, mode);
	restore_state(&setup);
	kgim_clock_set_mode(dpy, mode);
	kgim_graphics_set_mode(dpy, mode);

	chipset_setup_dpy(dpy, mode);

	NOTICE1("Mode set complete");
}

static int chipset_command(
	struct kgi_graphic *state, 
	kgi_uint cmd, 
	kgi_uint arg)
{
	kgi_coord  *coord;
	struct kgi_font *font;
	struct kgi_display *dpy;

	dpy = state->dev.dpy;
	switch (cmd) {
		case CHIP_SETDEFFONT:
		font = (struct kgi_font *)(state->io_buf);
		text16_set_font(dpy, font, 0, font->info->positions);
		return 0;
		case CHIP_SETSPLITLINE:
			if (arg > state->dev.mode.visible.y)
				return -E(CHIPSET,BOGUS);

 			if (state->dev.org[0].y > 
			    state->dev.mode.visible.y - arg )
			  kgi_set_origin(
					&(state->dev), 0, 
					state->dev.org[0].x,
					state->dev.mode.visible.y - arg );
			
			kgi_set_split(&(state->dev),arg*dpy->mode.magnify.y);
			return 0;
		case CHIP_SETVISFRAME:
			coord = (kgi_coord *)(state->io_buf);

			if (coord->x < 0 || 
			    coord->x > 
			    state->dev.mode.virt.x - state->dev.mode.visible.x)
				return -E(CHIPSET,BOGUS);

 			if (coord->y < 0 || dpy->mode.magnify.y<1 ||
 			    coord->y > state->dev.mode.virt.y 
			    - state->dev.split/dpy->mode.magnify.y)
				return -E(CHIPSET,BOGUS);

			kgi_set_origin(&(state->dev), 0, coord->x, coord->y);
			return 0;
		default:
			break;
	}

	return -E(CHIPSET,UNKNOWN);
}

int kgim_chipset_init(struct kgi_card *card,struct kgi_display *dpy)
{
	int err;

	if ((err = chipset_init(dpy)) != EOK) return err;
 
	strcpy(card->manufact, chipinfo.chipset.manufact);
	strcpy(card->model, chipinfo.chipset.model);
	card->modes = chipinfo.chipset.modes;
	card->flags = chipinfo.chipset.flags;
	card->maxdots = chipinfo.chipset.maxdots;

	card->chipset =& chipinfo.chipset;
	card->preset = default_kgi_mode;

	return EOK;
}

void kgim_chipset_done(struct kgi_display *dpy)
{
	chipset_done();
}

int  kgim_chipset_check_mode(
	struct kgi_display *dpy, 
	struct kgi_mode *mode,
	enum kgi_tm_cmd cmd)
{
	DEBUG1("Checking mode... (%dx%d %d)\n", mode->request.virt.x,
	       mode->request.virt.y, mode->request.graphtype);
	return chipset_check_mode(dpy, mode, cmd);
}

void kgim_chipset_set_mode(struct kgi_display *dpy, struct kgi_mode *mode)
{
  DEBUG1("Setting mode... (%dx%d %d)\n", 
				 mode->request.virt.x,
				 mode->request.virt.y, 
				 mode->request.graphtype);
	chipset_set_mode(dpy,mode);
}

int kgim_chipset_command(
	struct kgi_graphic *state, 
	kgi_uint cmd, 
	kgi_uint iobuf)
{
	return chipset_command(state, cmd, iobuf);
}
